@@ -17,12 +17,13 @@
 **
 ****************************************************************************/
 
-#include <stdlib.h>
+#include <cstdlib>
 
-#include <qsmsmessage.h>
+#include "qsmsmessage.h"
+#include "qcbsmessage.h"
 #include "qsmsmessage_p.h"
-#include <qatutils.h>
-#include <qgsmcodec.h>
+#include "qatutils.h"
+#include "qgsmcodec.h"
 #ifdef Q_WS_QWS
 #include <qtopialog.h>
 #else
@@ -84,7 +85,7 @@ public:
         mReplyRequest = false;
         mStatusReportRequested = false;
         mMessageType = QSMSMessage::Normal;
-        mCodec = 0;
+        mCodec = nullptr;
         mForceGsm = false;
         mBestScheme = (QSMSDataCodingScheme)(-1);
         mDataCodingScheme = -1;
@@ -364,7 +365,11 @@ QSMSMessage::~QSMSMessage()
 QSMSMessagePrivate *QSMSMessage::dwrite()
 {
     // If we are the only user of the private object, return it as-is.
-    if ( d->ref == 1 )
+#if (QT_VERSION >= QT_VERSION_CHECK(5, 14, 0))
+    if ( d->ref.loadRelaxed() == 1 )
+#else
+    if ( d->ref.load() == 1 )
+#endif
         return d;
 
     // Create a new private object and copy the current contents into it.
@@ -1560,7 +1565,7 @@ bool QPDUMessage::bit(int b)
     if ( needOctets(1) )
         return (mBuffer[mPosn] & (Unit << b));
     else
-        return 0;
+        return false;
 }
 
 unsigned char QPDUMessage::bits(int offset, int len)
@@ -2029,7 +2034,7 @@ QString QPDUMessage::userData(QSMSDataCodingScheme scheme, QTextCodec *codec, QB
     uint ch;
 
     // Reset the header return.
-    headers = 0;
+    headers = nullptr;
 
     // Get the length of the user data payload.
     if ( implicitLength ) {
@@ -2347,11 +2352,9 @@ static void unpackSckl( QSMSMessage& m, const QString& text )
 QSMSMessage QSMSDeliverMessage::unpack(QTextCodec *codec)
 {
     QSMSMessage m;
-    bool moreMessages;
     bool statusReport;
     bool userDataHeader;
     bool replyPath;
-    bool rejectDuplicates;
     unsigned char protocol;
     unsigned char scheme;
     uint msgType;
@@ -2370,16 +2373,12 @@ QSMSMessage QSMSDeliverMessage::unpack(QTextCodec *codec)
         return m;
     msgType = bits(0, 2);
     if ( msgType == SMS_Deliver ) {
-        moreMessages = bit(2);
         // Bits 3 and 4 are unused for deliver messages.
         statusReport = bit(5);
         userDataHeader = bit(6);
         replyPath = bit(7);
-        rejectDuplicates = false;
         validityFormat = SMS_VF_NoPresent;
     } else if ( msgType == SMS_Submit ) {
-        moreMessages = false;
-        rejectDuplicates = bit(2);
         validityFormat = bits(3, 2);
         statusReport = bit(5);
         userDataHeader = bit(6);
@@ -2442,7 +2441,7 @@ QSMSMessage QSMSDeliverMessage::unpack(QTextCodec *codec)
     }
 
     // Read the user data field.
-    QByteArray *headers = 0;
+    QByteArray *headers = nullptr;
     QString text;
     text = userData( (QSMSDataCodingScheme)scheme, codec,
                      headers, userDataHeader, false );
@@ -2468,6 +2467,93 @@ QSMSMessage QSMSDeliverMessage::unpack(QTextCodec *codec)
     return m;
 }
 
+static void iso639_2_from_language(QCBSMessage::Language lang, char *iso639)
+{
+    switch (lang) {
+    case QCBSMessage::German:
+        iso639[0] = 'd';
+        iso639[1] = 'e';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::English:
+        iso639[0] = 'e';
+        iso639[1] = 'n';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Italian:
+        iso639[0] = 'i';
+        iso639[1] = 't';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::French:
+        iso639[0] = 'f';
+        iso639[1] = 'r';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Spanish:
+        iso639[0] = 'e';
+        iso639[1] = 's';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Dutch:
+        iso639[0] = 'n';
+        iso639[1] = 'l';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Swedish:
+        iso639[0] = 's';
+        iso639[1] = 'v';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Danish:
+        iso639[0] = 'd';
+        iso639[1] = 'a';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Portuguese:
+        iso639[0] = 'p';
+        iso639[1] = 't';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Finnish:
+        iso639[0] = 'f';
+        iso639[1] = 'i';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Norwegian:
+        iso639[0] = 'n';
+        iso639[1] = 'o';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Greek:
+        iso639[0] = 'e';
+        iso639[1] = 'l';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Turkish:
+        iso639[0] = 't';
+        iso639[1] = 'r';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Hungarian:
+        iso639[0] = 'h';
+        iso639[1] = 'u';
+        iso639[2] = '\0';
+        break;
+    case QCBSMessage::Polish:
+        iso639[0] = 'p';
+        iso639[1] = 'l';
+        iso639[2] = '\0';
+        break;
+    default:
+        iso639[0] = '\0';
+        iso639[1] = '\0';
+        iso639[2] = '\0';
+        break;
+    }
+
+    return;
+}
 
 QCBSDeliverMessage::QCBSDeliverMessage()
     : QPDUMessage()
@@ -2508,12 +2594,12 @@ QCBSMessage QCBSDeliverMessage::unpack(QTextCodec *codec)
     mPosn += 6;
 
     // Read the user data field and strip CR's, LF's, and NUL's from the end.
-    QByteArray *headers = 0;
+    QByteArray *headers = nullptr;
     QString text = userData
         ( (QSMSDataCodingScheme)scheme, codec, headers, false, true );
     len = text.length();
-    while ( len > 0 && ( text[len - 1] == '\r' || text[len - 1] == '\n' ||
-                         text[len - 1] == '\0' ) ) {
+    while ( len > 0 && ( text.at(len - 1) == QChar ('\r') || text.at(len - 1) == QChar ('\n') ||
+                         text.at(len - 1) == QChar ('\0') ) ) {
         --len;
     }
     m.setText( text.left( len ) );
@@ -2524,29 +2610,44 @@ QCBSMessage QCBSDeliverMessage::unpack(QTextCodec *codec)
 
 void QCBSDeliverMessage::pack(const QCBSMessage &m, QSMSDataCodingScheme scheme)
 {
+    int numPad;
+    QTextCodec *codec = QAtUtils::codec( "gsm" );
+    QByteArray header;
+
     // Clear the pdu before we start.
     mBuffer = QByteArray();
     mPosn = 0;
     mBits = 0;
 
-    scheme = QSMS_DefaultAlphabet;
-    QByteArray data;
     mBuffer.append( (char) (((m.messageCode() >> 4) & 0x3F) | (m.scope() << 6)) );
     mBuffer.append( (char)(((m.messageCode() & 0xF) << 4) | (m.updateNumber() & 0xF)) );
     mBuffer.append( (char)((m.channel() & 0x0000FF00) >> 8) );
     mBuffer.append( (char)(m.channel() & 0x000000FF) );
-    mBuffer.append( (char)(((scheme & 0x0F)<<4) | (m.language() & 0x0F)) );
-    mBuffer.append( (char)((m.numPages() & 0x0F) | ((m.page() & 0x0F) << 4)) );
 
-    QTextCodec *codec = QAtUtils::codec( "gsm" );
-    QByteArray header;
+    if (scheme == QSMS_UCS2Alphabet)
+        mBuffer.append( 0x11 );
+    else
+        mBuffer.append( (char)(((scheme & 0x0F)<<4) | (m.language() & 0x0F)) );
+
+    mBuffer.append( (char)((m.numPages() & 0x0F) | ((m.page() & 0x0F) << 4)) );
 
     QString paddedText = m.text();
 
-    int numPad = 93 - getEncodedLength(paddedText, paddedText.length());
+    if (scheme == QSMS_UCS2Alphabet) {
+        char iso639[3];
+        unsigned short c;
+
+        iso639_2_from_language(m.language(), iso639);
+        c = ((iso639[1] & 0x7F)<<7) | (iso639[0] & 0x7F);
+        mBuffer.append( (char)(c & 0xFF) );
+        mBuffer.append( (char)(c >> 8) );
+        numPad = 40 - (paddedText.length());
+    }
+    else
+        numPad = 93 - getEncodedLength(paddedText, paddedText.length());
 
     for (int i = 0; i < numPad; i++)
-    	paddedText.append(QChar(0x0D));
+        paddedText.append(QChar(0x0D));
 
     setUserData(paddedText, scheme, codec, header,true);
 }
